// The necessary parts of ANSI escape sequence support required for this editor to work

const std = @import("std");
const fmt = std.fmt;
const Color = @import("color.zig").Color;
const Pos = @import("pos.zig");

pub fn parseCsiParams(res: []usize, comptime default: usize, params: []const u8, min_params: usize) ![]usize {
    var i: usize = 0;
    var si = std.mem.splitScalar(u8, params, ';');
    while (i < res.len) : (i += 1) {
        const param = si.next();
        if (param) |num| {
            res[i] = if (num.len == 0) default else try std.fmt.parseInt(usize, num, 10);
        } else {
            res[i] = default;
            break;
        }
    }
    if (i == res.len and si.next() != null)
        return error.TooManyParameters;
    while (i < min_params) : (i += 1)
        res[i] = default;
    return res[0..i];
}

pub const AnsiCommand = union(enum) {
    ignore, // something we don't recognize
    tab,
    backspace,
    delete,
    newline,
    carriage_return,
    insert,
    // Control characters (excluding those corresponding to the above)
    ctrl: u8, // a - z
    // Regular printable ASCII key presses
    print: u8,
    cursor_movement: CursorMovement,
    const CursorMovement = struct {
        const Action = enum { left, right, up, down, pos1, end };
        modifier: enum { none, shift, ctrl },
        action: Action,
        times: usize,
    };
};

pub const AnsiCommandIterator = struct {
    buf: []const u8,
    i: usize,
    fn getch(self: *AnsiCommandIterator) ?u8 {
        if (self.eof()) return null;
        const c = self.buf[self.i];
        self.i += 1;
        return c;
    }
    fn ungetch(self: *AnsiCommandIterator) void {
        self.i -= 1;
    }
    fn eof(self: *AnsiCommandIterator) bool {
        return self.i >= self.buf.len;
    }
    fn parseCursorMovement(
        kind: u8,
        params: []const u8,
    ) AnsiCommand {
        var movbuf: [2]usize = undefined;
        const movement = parseCsiParams(&movbuf, 1, params, 1) catch return .{ .ignore = {} };
        const action: AnsiCommand.CursorMovement.Action = switch (kind) {
            'A' => .up,
            'B' => .down,
            'C' => .right,
            'D' => .left,
            'H' => .pos1,
            'F' => .end,
            else => return .{ .ignore = {} },
        };
        const times = movement[0];
        return if (movement.len == 1)
            .{ .cursor_movement = .{ .modifier = .none, .action = action, .times = times } }
        else if (movement.len == 2)
            switch (movement[1]) {
                2 => .{ .cursor_movement = .{ .modifier = .shift, .action = action, .times = times } },
                5 => .{ .cursor_movement = .{ .modifier = .ctrl, .action = action, .times = times } },
                else => .{ .ignore = {} },
            }
        else
            .{ .ignore = {} };
    }
    pub fn next(self: *AnsiCommandIterator) ?AnsiCommand {
        const c = self.getch() orelse return null;
        return switch (c) {
            else => .{ .ignore = {} },
            0x08 => .{ .backspace = {} },
            0x7F => .{ .delete = {} },
            '\n' => .{ .newline = {} },
            '\r' => .{ .carriage_return = {} },
            '\t' => .{ .tab = {} },
            0x01...0x07, 0x0C, 0x0E...0x1A => .{ .ctrl = c - 1 + 'a' },
            // Printable ASCII
            ' '...'~' => .{ .print = c },
            // ANSI escape sequences
            0x1B => switch (self.getch() orelse return .{ .ignore = {} }) { // TODO possible ESC
                else => .{ .ignore = {} },
                // Fe sequence
                0x40...0x5A, 0x5C...0x5F => .{ .ignore = {} },
                '[' => {
                    // CSI sequence
                    if (self.eof()) return .{ .ignore = {} }; // TODO cut off?
                    const start = self.i;
                    var c2 = self.getch() orelse return .{ .ignore = {} };
                    while (c2 >= 0x30 and c2 <= 0x3F)
                        c2 = self.getch() orelse return .{ .ignore = {} };
                    while (c2 >= 0x20 and c2 <= 0x2F)
                        c2 = self.getch() orelse return .{ .ignore = {} };
                    if (c2 < 0x40 or c2 > 0x7E) {
                        self.ungetch();
                        return .{ .ignore = {} }; // TODO syntax error
                    }
                    const params = self.buf[start .. self.i - 1];
                    if (c2 == '~') return switch (fmt.parseInt(usize, params, 10) catch return .{ .ignore = {} }) {
                        2 => .{ .insert = {} },
                        else => .{ .ignore = {} },
                    };
                    return parseCursorMovement(c2, params);
                },
            },
        };
    }
};

pub fn parse(buf: []const u8) AnsiCommandIterator {
    return .{ .buf = buf, .i = 0 };
}

pub fn writeCsi(writer: anytype, seq: []const u8) !void {
    try writer.writeAll("\x1B[");
    try writer.writeAll(seq);
}

pub fn fmtCsi(writer: anytype, csi_type: u8, params: anytype) !void {
    try writer.writeAll("\x1B[");
    if (params.len > 0)
        try fmt.format(writer, "{d}", .{params[0]});
    var i: usize = 1;
    while (i < params.len) : (i += 1)
        try fmt.format(writer, ";{d}", .{params[i]});
    try writer.writeByte(csi_type);
}

pub fn setCursor(writer: anytype, pos: Pos) !void {
    return fmtCsi(writer, 'H', [_]usize{ pos.row + 1, pos.col + 1 });
}

pub fn setFg(writer: anytype, color: Color) !void {
    return fmtCsi(writer, 'm', [_]usize{ 38, 2, color.r, color.g, color.b });
}

pub fn setBg(writer: anytype, color: Color) !void {
    return fmtCsi(writer, 'm', [_]usize{ 48, 2, color.r, color.g, color.b });
}

pub fn clearScreen(writer: anytype) !void {
    return writeCsi(writer, "2J");
}
