const std = @import("std");
const os = std.os;
const posix = std.posix;
const File = std.fs.File;
const ArrayList = std.ArrayList;

const ansi = @import("ansi.zig");
const AnsiCommand = ansi.AnsiCommand;
const txtbuf = @import("txtbuf.zig");
const color = @import("color.zig");
const util = @import("util.zig");

const Pos = @import("pos.zig");
const TextBuffer = txtbuf.TextBuffer;

const tok = @import("tok.zig");

fn asgn(u: usize) isize {
    return @intCast(u);
}

fn asun(i: isize) usize {
    return @intCast(i);
}

const Allocator = std.mem.Allocator;

const Edit = struct {
    const Self = @This();
    at: usize,
    old: []const u8,
    new: []const u8,
    fn init(at: usize, old: []const u8, new: []const u8, a: Allocator) !Self {
        const self_new = try a.alloc(u8, new.len);
        @memcpy(self_new, new);
        const self_old = try a.alloc(u8, old.len);
        @memcpy(self_old, old);
        return .{ .at = at, .new = self_new, .old = self_old };
    }
    // Note: To play nice with arena allocators, this must be in reverse order vs. init!
    fn deinit(self: Self, a: Allocator) void {
        a.free(self.old);
        a.free(self.new);
    }
    fn redo(self: Self, buf: *TextBuffer) !usize {
        try buf.replaceRange(self.at, self.old.len, self.new);
        return self.at + self.new.len;
    }
    fn undo(self: Self, buf: *TextBuffer) !usize {
        try buf.replaceRange(self.at, self.new.len, self.old);
        return self.at + self.old.len;
    }
};

const allocator = std.heap.c_allocator;

// TODO cap history size

const History = struct {
    const Self = @This();
    const ArenaAllocator = std.heap.ArenaAllocator;
    arena: ArenaAllocator,
    edits: ArrayList(Edit),
    cursor: usize,
    pub fn init() Self {
        const arena = ArenaAllocator.init(std.heap.page_allocator);
        return .{
            .arena = arena,
            .edits = ArrayList(Edit).init(allocator),
            .cursor = 0,
        };
    }
    pub fn deinit(self: Self) void {
        self.arena.deinit();
        self.edits.deinit();
    }

    fn undo(self: *Self, text: *TextBuffer) !?usize {
        if (self.cursor == 0)
            return null;
        self.cursor -= 1;
        return try self.edits.items[self.cursor].undo(text);
    }

    fn redo(self: *Self, text: *TextBuffer) !?usize {
        if (self.cursor == self.edits.items.len)
            return null;
        const res = try self.edits.items[self.cursor].redo(text);
        self.cursor += 1;
        return res;
    }

    fn record(self: *Self, at: usize, old: []const u8, new: []const u8) !void {
        // TODO do better than throwing edits away
        // Deinitialize *in reverse order* so the arena can actually free the memory
        if (self.cursor < self.edits.items.len) {
            var i: usize = self.edits.items.len - 1;
            while (true) {
                self.edits.items[i].deinit(self.arena.allocator());
                if (i == self.cursor) break;
                i -= 1;
            }
            self.edits.shrinkRetainingCapacity(self.cursor);
        }
        try self.edits.append(try Edit.init(at, old, new, self.arena.allocator()));
        self.cursor += 1;
    }
};

const Clipboard = struct {
    const Self = @This();
    content: []u8,
    pub fn init() !Self {
        return .{ .content = try allocator.alloc(u8, 0) };
    }
    pub fn deinit(self: Self) void {
        allocator.free(self.content);
    }
    pub fn get(self: Self) []const u8 {
        return self.content;
    }
    pub fn set(self: *Self, content: []const u8) !void {
        self.content = try allocator.realloc(self.content, content.len);
        @memcpy(self.content, content);
    }
};

const TextPane = struct {
    const Self = @This();
    anchor: Pos,
    cursor: Pos,
    mark: ?Pos,
    text: TextBuffer,
    history: History,
    clipboard: *Clipboard,
    multiline: bool,
    pub fn init(clipboard: *Clipboard, multiline: bool) !Self {
        return .{
            .anchor = .{ .row = 0, .col = 0 },
            .cursor = .{ .row = 0, .col = 0 },
            .mark = null,
            .text = try TextBuffer.empty(),
            .history = History.init(),
            .clipboard = clipboard,
            .multiline = multiline,
        };
    }
    pub fn deinit(self: *Self) void {
        self.text.deinit();
        self.history.deinit();
    }

    fn setCursor(self: *Self, cursor: Pos) void {
        self.cursor = cursor;
        self.mark = null;
    }

    pub fn setText(self: *Self, text: TextBuffer) void {
        self.text = text;
        self.setCursor(.{ .row = 0, .col = 0 });
        self.anchor = .{ .row = 0, .col = 0 };
        self.history.deinit();
        self.history = History.init();
    }

    fn moveRows(self: *Self, d: isize) Pos {
        const row = @max(0, @min(asgn(self.text.nRows()) - 1, asgn(self.cursor.row) + d));
        const col = @max(0, @min(self.text.nCols(row), self.cursor.col));
        return .{ .row = row, .col = col };
    }

    fn moveCols(self: *Self, d_cols: isize) Pos {
        var cursor = self.cursor;
        var d = d_cols;
        while (d < 0) {
            const col = asgn(cursor.col) + d;
            if (col >= 0) {
                cursor.col = asun(col);
                break;
            }
            if (cursor.row == 0) {
                cursor.col = 0;
                break;
            }
            cursor.row -= 1;
            cursor.col = self.text.nCols(cursor.row);
            d = col + 1;
        }
        while (d > 0) {
            const col = asgn(cursor.col) + d;
            const nCols = self.text.nCols(cursor.row);
            if (col <= nCols) {
                cursor.col = asun(col);
                break;
            }
            if (cursor.row == self.text.nRows() - 1) {
                cursor.col = nCols;
                break;
            }
            cursor.row += 1;
            cursor.col = 0;
            d = col - asgn(nCols) - 1;
        }
        return cursor;
    }

    fn getCursorIdx(self: Self) usize {
        return self.text.getIdx(self.cursor);
    }

    const Span = struct { min: usize, max: usize };
    fn getSelection(self: Self) ?Span {
        if (self.mark) |mark| {
            const idx = self.getCursorIdx();
            const mark_idx = self.text.getIdx(mark);
            return .{ .min = @min(idx, mark_idx), .max = @max(idx, mark_idx) };
        }
        return null;
    }

    fn getCopySpan(self: Self) Span {
        const idx = self.text.getIdx(.{ .row = self.cursor.row, .col = 0 });
        return self.getSelection() orelse Span{ .min = idx, .max = idx + self.text.nCols(self.cursor.row) + 1 };
    }

    fn copy(self: *Self) !void {
        const sel = self.getCopySpan();
        try self.clipboard.set(self.text.getRange(sel.min, sel.max));
    }

    fn deleteSpan(self: *Self, sel: Span) !void {
        const old = self.text.getRange(sel.min, sel.max);
        try self.history.record(sel.min, old, &.{});
        try self.text.replaceRange(sel.min, old.len, &.{});
        self.setCursor(self.text.getPos(sel.min));
    }

    fn delete(self: *Self) !void {
        const idx = self.getCursorIdx();
        const sel = self.getSelection() orelse Span{ .min = if (idx == 0) return else idx - 1, .max = idx };
        return self.deleteSpan(sel);
    }

    fn cut(self: *Self) !void {
        try self.copy();
        try self.deleteSpan(self.getCopySpan());
    }

    fn insert(self: *Self, new: []const u8) !void {
        if (!self.multiline and std.mem.indexOfScalar(u8, new, '\n') != null)
            return;
        const cur_idx = self.getCursorIdx();
        const sel = self.getSelection() orelse Span{ .min = cur_idx, .max = cur_idx };
        const old = self.text.getRange(sel.min, sel.max);
        try self.history.record(sel.min, old, new);
        try self.text.replaceRange(sel.min, old.len, new);
        self.setCursor(self.text.getPos(sel.min + new.len));
    }

    fn insertNewline(self: *Self) !void {
        const start_row = if (self.mark) |pos| @min(pos.row, self.cursor.row) else self.cursor.row;
        const start_col = if (start_row < self.cursor.row) self.mark.?.col else @min((self.mark orelse self.cursor).col, self.cursor.col);
        const start_idx = self.text.getIdx(.{ .row = start_row, .col = 0 });
        const line = self.text.getRange(start_idx, start_idx + self.text.nCols(start_row));
        var i: usize = 0;
        while (i < start_col and (line[i] == ' ' or line[i] == '\t')) i += 1;
        const indent = line[0..i];
        const indented_line = try std.mem.concat(allocator, u8, &.{ "\n", indent });
        defer allocator.free(indented_line);
        try self.insert(indented_line);
    }

    fn paste(self: *Self) !void {
        try self.insert(self.clipboard.get());
    }

    fn undo(self: *Self) !void {
        if (try self.history.undo(&self.text)) |idx|
            self.setCursor(self.text.getPos(idx));
    }

    fn redo(self: *Self) !void {
        if (try self.history.redo(&self.text)) |idx|
            self.setCursor(self.text.getPos(idx));
    }

    fn selectAll(self: *Self) void {
        self.cursor = .{ .row = 0, .col = 0 };
        self.mark = .{ .row = self.text.nRows() - 1, .col = self.text.nCols(self.text.nRows() - 1) };
    }

    fn selectToken(self: *Self) void {
        var hint = tok.Tokenization.Hint{};
        const token = self.text.tokens.getToken(self.getCursorIdx(), &hint);
        // HACK don't let the trailing newline disappear
        self.cursor = self.text.getPos(@min(hint.sum, self.text.chars.items.len - 1));
        self.mark = self.text.getPos(hint.sum - token.len);
    }

    fn gotoPrevToken(self: *Self) void {
        self.setCursor(self.text.getPos(self.text.tokens.getPreviousTokenStart(self.getCursorIdx())));
    }

    fn gotoNextToken(self: *Self) void {
        self.setCursor(self.text.getPos(self.text.tokens.getNextTokenEnd(self.getCursorIdx())));
    }

    pub fn processCommand(self: *Self, cmd: AnsiCommand) !void {
        switch (cmd) {
            else => {},
            .backspace, .delete => try self.delete(),
            .newline, .carriage_return => try self.insertNewline(),
            .print => |c| try self.insert(&[_]u8{c}),
            .tab => try self.insert("    "), // HACK awful
            .ctrl => |c| switch (c) {
                'z' => try self.undo(),
                'y' => try self.redo(),
                'c' => try self.copy(),
                'x' => try self.cut(),
                'v' => try self.paste(),
                'a' => self.selectAll(),
                't' => self.selectToken(),
                else => {},
            },
            .cursor_movement => |cm| switch (cm.modifier) {
                .none => switch (cm.action) {
                    .up => self.setCursor(self.moveRows(-asgn(cm.times))),
                    .down => self.setCursor(self.moveRows(asgn(cm.times))),
                    .right => self.setCursor(self.moveCols(asgn(cm.times))),
                    .left => self.setCursor(self.moveCols(-asgn(cm.times))),
                    .pos1 => self.setCursor(.{ .col = 0, .row = self.cursor.row }),
                    .end => self.setCursor(.{ .col = self.text.nCols(self.cursor.row), .row = self.cursor.row }),
                },
                .ctrl => switch (cm.action) {
                    .up, .down => {}, // TODO
                    .left => self.gotoPrevToken(),
                    .right => self.gotoNextToken(),
                    .pos1 => self.setCursor(.{ .row = 0, .col = 0 }),
                    .end => self.setCursor(.{ .row = self.text.nRows() - 1, .col = self.text.nCols(self.text.nRows() - 1) }),
                },
                .shift => {
                    if (self.mark == null)
                        self.mark = self.cursor;
                    switch (cm.action) {
                        .up => self.cursor = self.moveRows(-asgn(cm.times)),
                        .down => self.cursor = self.moveRows(asgn(cm.times)),
                        .right => self.cursor = self.moveCols(asgn(cm.times)),
                        .left => self.cursor = self.moveCols(-asgn(cm.times)),
                        .pos1 => self.cursor.col = 0,
                        .end => self.cursor.col = self.text.nCols(self.cursor.row),
                    }
                },
            },
        }
    }

    fn scroll(self: *Self, dim: Pos) void {
        self.anchor = self.anchor.min(self.cursor);
        const bottom_right = self.anchor.add(dim);
        if (self.cursor.row >= bottom_right.row)
            self.anchor.row += self.cursor.row - bottom_right.row + 1;
        if (self.cursor.col >= bottom_right.col)
            self.anchor.col += self.cursor.col - bottom_right.col + 1;
    }

    const ColorManager = struct {
        const CM = @This();
        fg: color.Color,
        bg: color.Color,
        pub fn init(writer: anytype) !CM {
            const fg = color.FG[@intFromEnum(tok.TokenType.ignore)];
            try ansi.setFg(writer, fg);
            const bg = color.BG;
            try ansi.setBg(writer, bg);
            return .{
                .fg = fg,
                .bg = bg,
            };
        }
        pub fn set(self: *CM, writer: anytype, t: tok.TokenType, sel: bool) !void {
            var fg = color.FG[@intFromEnum(t)];
            var bg = color.BG;
            if (sel)
                std.mem.swap(color.Color, &fg, &bg);
            if (!std.meta.eql(fg, self.fg)) {
                try ansi.setFg(writer, fg);
                self.fg = fg;
            }
            if (!std.meta.eql(bg, self.bg)) {
                try ansi.setBg(writer, bg);
                self.bg = bg;
            }
        }
    };

    fn render(self: *Self, dim: Pos, writer: anytype, focus: bool) !void {
        self.scroll(dim);

        const idx = self.getCursorIdx();
        const sel = self.getSelection() orelse Span{ .min = idx, .max = idx + 1 };

        var row: usize = 0;
        var cm = try ColorManager.init(writer);
        var tok_hint: tok.Tokenization.Hint = .{};
        while (row < dim.row) : (row += 1) {
            var col: usize = 0;
            while (col < dim.col) : (col += 1) {
                const pos = .{ .row = row, .col = col };
                const abs = self.anchor.add(pos);
                const c = self.text.get(abs);
                var highlight = false;
                var tt = tok.TokenType.ignore;
                if (abs.row < self.text.nRows() and abs.col <= self.text.nCols(abs.row)) {
                    const c_idx = self.text.getIdx(abs);
                    if (focus) highlight = c_idx >= sel.min and c_idx < sel.max;
                    tt = self.text.tokens.getToken(c_idx, &tok_hint).kind;
                }
                try cm.set(writer, tt, highlight);
                switch (c orelse ' ') {
                    0x20...0x7e => |ch| try writer.writeByte(ch),
                    else => try writer.writeAll(" "),
                }
            }
            if (row + 1 < dim.row)
                try writer.writeAll("\r\n");
        }
        try cm.set(writer, .ignore, false);
    }
};

pub const Editor = struct {
    const Self = @This();
    const BW = @TypeOf(std.io.bufferedWriter(std.io.getStdOut().writer()));
    tty: File,
    tui: *BW,
    path: []u8,
    clipboard: *Clipboard,
    textpane: TextPane,
    state: union(enum) {
        normal: void,
        confirm_quit: void,
        save_path: TextPane,
    },
    pub fn init(tty: File, tui: *BW) !Self {
        const clipboard = try allocator.create(Clipboard);
        clipboard.* = try Clipboard.init();
        return .{
            .tty = tty,
            .tui = tui,
            .path = try allocator.alloc(u8, 0),
            .clipboard = clipboard,
            .textpane = try TextPane.init(clipboard, true),
            .state = .normal,
        };
    }
    pub fn deinit(self: *Self) void {
        self.textpane.deinit();
        self.clipboard.deinit();
        allocator.free(self.path);
        allocator.destroy(self.clipboard);
    }

    fn setPath(self: *Self, path: []const u8) !void {
        self.path = try allocator.realloc(self.path, path.len);
        @memcpy(self.path, path);
    }

    pub fn open(self: *Self, path: []const u8) !void {
        if (TextBuffer.open(path)) |text| {
            self.textpane.setText(text);
            try self.setPath(path);
        } else |err| switch (err) {
            error.FileNotFound => {
                self.textpane.setText(try TextBuffer.empty());
                try self.setPath(path);
            },
            else => try self.setPath(""),
        }
    }

    fn isSaved(self: *Self) bool {
        if (self.path.len == 0)
            return self.textpane.text.chars.items.len == 0;
        const file = std.fs.cwd().openFile(self.path, .{ .mode = .read_only }) catch return false;
        var chars = ArrayList(u8).init(allocator);
        defer chars.deinit();
        file.reader().readAllArrayList(&chars, 16384) catch return false;
        return std.mem.eql(u8, self.textpane.text.chars.items, chars.items);
    }

    fn save(self: *Self) !void {
        if (self.path.len > 0) {
            try self.textpane.text.save(self.path); // TODO error handling
        } else {
            self.state = .{ .save_path = try TextPane.init(self.clipboard, false) };
        }
    }

    fn processCommand(self: *Self, cmd: AnsiCommand) !bool {
        switch (self.state) {
            .normal => switch (cmd) {
                .ctrl => |c| switch (c) {
                    'q' => {
                        if (self.isSaved())
                            return true;
                        self.state = .{ .confirm_quit = {} };
                    },
                    's' => try self.save(),
                    else => try self.textpane.processCommand(cmd),
                },
                else => try self.textpane.processCommand(cmd),
            },
            .confirm_quit => switch (cmd) {
                .ctrl => |c| switch (c) {
                    'q' => return true,
                    else => self.state = .{ .normal = {} },
                },
                else => self.state = .{ .normal = {} },
            },
            .save_path => |prompt| switch (cmd) {
                .ctrl => |c| switch (c) {
                    'q' => {
                        self.state.save_path.deinit();
                        self.state = .{ .normal = {} };
                    },
                    else => try self.state.save_path.processCommand(cmd),
                },
                .newline, .carriage_return => {
                    const path = prompt.text.chars.items;
                    try self.setPath(path[0 .. path.len - 1]); // trim trailing newline (TODO refucktor)
                    try self.textpane.text.save(self.path); // TODO error handling
                    self.state.save_path.deinit();
                    self.state = .{ .normal = {} };
                },
                else => try self.state.save_path.processCommand(cmd),
            },
        }
        return false;
    }

    fn processInput(self: *Self, buf: []const u8) !bool {
        var iter = ansi.parse(buf);
        while (iter.next()) |cmd| {
            switch (cmd) {
                .ignore => {},
                .insert => @import("e/josh.zig").josh(self.tty, self.tui) catch {},
                else => if (try self.processCommand(cmd)) return true,
            }
        }
        return false;
    }

    fn render(self: *Self) !void {
        const us = std.time.microTimestamp();
        const writer = self.tui.writer();
        const dim = try util.getTtySize(self.tty);
        try ansi.setCursor(writer, .{ .row = 0, .col = 0 });
        try self.textpane.render(.{ .row = dim.row - 1, .col = dim.col }, writer, self.state == .normal);
        try writer.writeAll("\r\n");

        switch (self.state) {
            .normal => {
                var buf: [1000]u8 = undefined;
                var fixbuf = std.io.fixedBufferStream(&buf);
                @memset(buf[0..dim.col], ' ');
                std.fmt.format(fixbuf.writer(), "{s}:{d}:{d}  {d} us  {d} lines", .{
                    if (self.path.len == 0) "?" else self.path,
                    self.textpane.cursor.row + 1,
                    self.textpane.cursor.col + 1,
                    std.time.microTimestamp() - us,
                    self.textpane.text.nRows(),
                }) catch {};
                try writer.writeAll(buf[0..dim.col]);
            },
            .confirm_quit => {
                const msg = "You have unsaved changes. Press C-q to quit, any other key to cancel.";
                try writer.writeAll(if (dim.col < msg.len) msg[0..dim.col] else msg);
                var i = msg.len;
                while (i < dim.col) : (i += 1) try writer.writeByte(' ');
            },
            .save_path => {
                const prefix = "save as: ";
                try writer.writeAll(prefix);
                try self.state.save_path.render(.{ .row = 1, .col = dim.col - prefix.len }, writer, true);
            },
        }

        try self.tui.flush();
    }

    pub fn run(self: *Self) !void {
        try self.render();
        var buf: [1000]u8 = undefined;
        while (true) {
            // TODO deal with full buffer (unlikely in normal usage)
            const len = try self.tty.read(&buf);
            if (try self.processInput(buf[0..len]))
                return;
            try self.render();
        }
    }
};
