const std = @import("std");
const Allocator = std.mem.Allocator;
const ArrayList = std.ArrayListUnmanaged;
const DynamicBitSet = std.bit_set.DynamicBitSet;

const AutoHashMap = std.AutoArrayHashMapUnmanaged;
fn CharMap(V: type) type {
    return AutoHashMap(u8, V);
}

pub fn DFA(V: type) type {
    const State = struct {
        const Self = @This();
        out: CharMap(usize),
        accepting: bool,
        value: V,
        pub fn init(allocator: Allocator, accepting: bool, value: V) !Self {
            return .{
                .out = try AutoHashMap(u8, usize).init(allocator, &.{}, &.{}),
                .accepting = accepting,
                .value = value,
            };
        }
        pub fn deinit(self: *Self, allocator: Allocator) void {
            self.out.deinit(allocator);
        }
    };
    return struct {
        const Self = @This();
        states: ArrayList(State),
        allocator: Allocator,
        pub fn init(allocator: Allocator) Self {
            return .{
                .states = ArrayList(State).initBuffer(&.{}),
                .allocator = allocator,
            };
        }
        pub fn deinit(self: *Self) void {
            for (self.states.items) |*state|
                state.deinit(self.allocator);
            self.states.deinit(self.allocator);
        }
        pub fn addState(self: *Self, accepting: bool, value: V) !usize {
            try self.states.append(self.allocator, try State.init(self.allocator, accepting, value));
            return self.states.items.len - 1;
        }
        pub fn addTransition(self: *Self, from: usize, via: u8, to: usize) !void {
            try self.states.items[from].out.put(self.allocator, via, to);
        }
        pub fn getTransition(self: Self, from: usize, via: u8) ?usize {
            return self.states.items[from].out.get(via);
        }
        pub fn accepts(self: Self, input: []const u8) bool {
            var state: usize = 0;
            for (input) |c| {
                if (self.getTransition(state, c)) |next_state|
                    state = next_state
                else
                    return false;
            }
            return self.states.items[state].accepting;
        }
    };
}
