const std = @import("std");
const Allocator = std.mem.Allocator;
const NFAC = @import("nfa.zig").NFA;
const DFAC = @import("dfa.zig").DFA;
const CharSet = std.bit_set.IntegerBitSet(256);

pub fn RegexReader(V: type) type {
    return struct {
        const NFA = NFAC(V);
        const DFA = DFAC(V);
        const Self = @This();
        pub const SyntaxError = error{
            UnexpectedEOF,
            UnclosedParenthesis,
            ExpectedOperand,
            UnexpectedChar,
            InvalidEscape,
        };
        pub const ReadError = SyntaxError || Allocator.Error;
        nfa: *NFA,
        i: usize,
        buf: []const u8,
        pub fn init(nfa: *NFA, expr: []const u8) Self {
            return .{
                .nfa = nfa,
                .i = 0,
                .buf = expr,
            };
        }
        fn getch(self: *Self) ?u8 {
            if (self.eof()) return null;
            const c = self.buf[self.i];
            self.i += 1;
            return c;
        }
        fn getch2(self: *Self) ReadError!u8 {
            return self.getch() orelse return SyntaxError.UnexpectedEOF;
        }
        fn ungetch(self: *Self) void {
            self.i -= 1;
        }
        fn eof(self: Self) bool {
            return self.i >= self.buf.len;
        }
        fn readEscape(self: *Self) ReadError!u8 {
            return switch (try self.getch2()) {
                'x' => l: {
                    const hi = try self.getch2();
                    const lo = try self.getch2();
                    break :l std.fmt.parseUnsigned(u8, &.{ hi, lo }, 16) catch |e| switch (e) {
                        error.Overflow => unreachable,
                        error.InvalidCharacter => return SyntaxError.InvalidEscape,
                    };
                },
                else => |c| c,
            };
        }
        fn readAtom(self: *Self) ReadError!?NFA.Sub {
            if (self.eof()) return null;
            switch (self.getch().?) {
                '(' => {
                    const res = try self.readChoice();
                    return if (try self.getch2() != ')') SyntaxError.UnclosedParenthesis else res;
                },
                '%' => return try self.nfa.str(&.{try self.readEscape()}),
                '[' => {
                    var charset = CharSet.initEmpty();
                    const complement = try self.getch2() == '^';
                    if (!complement) self.ungetch();
                    while (true) {
                        var c = try self.getch2();
                        if (c == ']') break;
                        if (c == '%')
                            c = try self.readEscape();
                        const c2 = try self.getch2();
                        if (c2 == '-') {
                            const c3 = try self.getch2();
                            if (c > c3) continue; // empty range
                            charset.setRangeValue(.{ .start = c, .end = c3 }, true);
                        } else {
                            charset.set(c);
                            self.ungetch();
                        }
                    }
                    if (complement) charset = charset.complement();
                    return try self.nfa.charset(charset);
                },
                ')', '?', '+', '*', '|' => {
                    self.ungetch();
                    return null;
                },
                else => |c| return try self.nfa.str(&.{c}),
            }
        }
        fn readQuantified(self: *Self) ReadError!?NFA.Sub {
            const base = try self.readAtom() orelse return null;
            return try switch (self.getch() orelse return base) {
                '?' => self.nfa.opt(base),
                '+' => self.nfa.rep(base),
                '*' => self.nfa.kleene(base),
                else => {
                    self.ungetch();
                    return base;
                },
            };
        }
        fn readConcat(self: *Self) ReadError!NFA.Sub {
            var base = try self.readQuantified() orelse return SyntaxError.ExpectedOperand;
            while (try self.readQuantified()) |q| base = try self.nfa.concat(&.{ base, q });
            return base;
        }
        fn readChoice(self: *Self) ReadError!NFA.Sub {
            var base = try self.readConcat();
            while (true) {
                if (self.getch()) |c| {
                    if (c == '|') base = try self.nfa.choice(&.{ base, try self.readConcat() }) else {
                        self.ungetch();
                        return base;
                    }
                } else return base;
            }
        }
        pub fn read(self: *Self) !NFA.Sub {
            const res = try self.readChoice();
            if (!self.eof())
                return error.UnexpectedChar;
            return res;
        }
    };
}

fn nothing(_: void, _: void) void {
    return {};
}

pub fn regexToDFA(allocator: Allocator, regex: []const u8) !DFAC(void) {
    var nfa = NFAC(void).init(allocator, {});
    defer nfa.deinit();
    var rr = RegexReader(void).init(&nfa, regex);
    const sub = try rr.read();
    return nfa.toDFA(sub, nothing);
}
