const std = @import("std");
const Allocator = std.mem.Allocator;
const CharSet = std.bit_set.IntegerBitSet(256);
fn HashMap(K: type, V: type) type {
    return std.ArrayHashMapUnmanaged(K, V, struct {
        pub fn hash(_: @This(), k: K) u32 {
            var hasher = std.hash.Wyhash.init(1233543646);
            std.hash.autoHashStrat(&hasher, k, .DeepRecursive);
            return @truncate(hasher.final());
        }
        pub fn eql(_: anytype, k1: K, k2: K, _: anytype) bool {
            return std.mem.eql(@typeInfo(K).Pointer.child, k1, k2);
        }
    }, true);
}
const AutoHashMap = std.AutoArrayHashMapUnmanaged;
const ArrayList = std.ArrayListUnmanaged;

const DFA = @import("dfa.zig").DFA;

const Subset = []const usize;

fn orderUsize(_: void, a: usize, b: usize) std.math.Order {
    return std.math.order(a, b);
}

fn contains(set: Subset, end_state: usize) bool {
    return std.sort.binarySearch(usize, end_state, set, {}, orderUsize) != null;
}

pub fn NFA(V: type) type {
    const Transition = struct { char: ?u8, to: usize };

    const State = struct {
        const Self = @This();

        out: ArrayList(Transition),
        value: V,
        pub fn init(value: V) Self {
            return .{
                .out = ArrayList(Transition).initBuffer(&.{}),
                .value = value,
            };
        }
        pub fn deinit(self: *Self, allocator: Allocator) void {
            self.out.deinit(allocator);
        }
    };

    return struct {
        const Self = @This();

        pub const Sub = struct {
            start: usize,
            end: usize,
        };

        allocator: Allocator,
        states: ArrayList(State),
        value: V, // HACK this is lazy

        pub fn init(allocator: Allocator, value: V) Self {
            return .{
                .allocator = allocator,
                .states = ArrayList(State).initBuffer(&.{}),
                .value = value,
            };
        }
        pub fn deinit(self: *Self) void {
            for (self.states.items) |*state|
                state.deinit(self.allocator);
            self.states.deinit(self.allocator);
        }
        pub fn addState(self: *Self) Allocator.Error!usize {
            try self.states.append(self.allocator, State.init(self.value));
            return self.states.items.len - 1;
        }
        pub fn addTransition(self: *Self, from: usize, char: ?u8, to: usize) Allocator.Error!void {
            try self.states.items[from].out.append(self.allocator, .{ .char = char, .to = to });
        }
        pub fn rep(self: *Self, sub: Sub) Allocator.Error!Sub {
            try self.addTransition(sub.end, null, sub.start);
            return sub;
        }
        pub fn opt(self: *Self, sub: Sub) Allocator.Error!Sub {
            try self.addTransition(sub.start, null, sub.end);
            return sub;
        }
        pub fn kleene(self: *Self, sub: Sub) Allocator.Error!Sub {
            return self.rep(try self.opt(sub));
        }
        pub fn choice(self: *Self, subs: []const Sub) Allocator.Error!Sub {
            const new = Sub{ .start = try self.addState(), .end = try self.addState() };
            for (subs) |sub| {
                try self.addTransition(new.start, null, sub.start);
                try self.addTransition(sub.end, null, new.end);
            }
            return new;
        }
        pub fn concat(self: *Self, subs: []const Sub) Allocator.Error!Sub {
            var last_end = subs[0].end;
            for (subs[1..]) |sub| {
                try self.addTransition(last_end, null, sub.start);
                last_end = sub.end;
            }
            return .{ .start = subs[0].start, .end = last_end };
        }
        pub fn str(self: *Self, s: []const u8) Allocator.Error!Sub {
            const start = try self.addState();
            var end = start;
            for (s) |c| {
                const new_end = try self.addState();
                try self.addTransition(end, c, new_end);
                end = new_end;
            }
            return .{ .start = start, .end = end };
        }
        pub fn charset(self: *Self, set: CharSet) Allocator.Error!Sub {
            const start = try self.addState();
            const end = try self.addState();
            var chars = set.iterator(.{});
            while (chars.next()) |c|
                try self.addTransition(start, @intCast(c), end);
            return .{ .start = start, .end = end };
        }

        fn hull(self: Self, states: Subset, hull_allocator: Allocator) Allocator.Error![]usize {
            const allocator = self.allocator; // should maybe be separate (e.g. arena allocation while algo runs) but meh
            var level = try ArrayList(usize).initCapacity(allocator, 1);
            defer level.deinit(allocator);
            try level.appendSlice(allocator, states);
            var next_level = ArrayList(usize).initBuffer(&.{});
            defer next_level.deinit(allocator);
            var seen = try AutoHashMap(usize, void).init(allocator, states, allocator.alloc(void, states.len) catch unreachable);
            defer seen.deinit(allocator);
            while (level.items.len > 0) {
                for (level.items) |state| {
                    for (self.states.items[state].out.items) |transition| {
                        if (transition.char == null and seen.get(transition.to) == null) {
                            try seen.put(allocator, transition.to, {});
                            try next_level.append(allocator, transition.to);
                        }
                    }
                }
                level.clearRetainingCapacity();
                const old_level = level;
                level = next_level;
                next_level = old_level;
            }
            const res = try hull_allocator.alloc(usize, seen.count());
            var iter = seen.iterator();
            var i: usize = 0;
            while (iter.next()) |e| {
                res[i] = e.key_ptr.*;
                i += 1;
            }
            return res;
        }
        pub fn toDFA(self: Self, sub: Sub, merge_values: fn (v1: V, v2: V) V) Allocator.Error!DFA(V) {
            const allocator = self.allocator; // same here
            var dfa = DFA(V).init(allocator);
            var subsets = std.heap.ArenaAllocator.init(allocator);
            defer subsets.deinit();
            const start_subset = try self.hull(&.{sub.start}, subsets.allocator());
            var nfa_to_dfa = try HashMap(Subset, usize).init(allocator, &.{start_subset}, &.{
                try dfa.addState(contains(start_subset, sub.end), self.states.items[sub.start].value),
            });
            defer nfa_to_dfa.deinit(allocator);
            var next_level = ArrayList(Subset).initBuffer(&.{});
            defer next_level.deinit(allocator);
            var level = try ArrayList(Subset).initCapacity(allocator, 1);
            defer level.deinit(allocator);
            try level.append(allocator, start_subset);
            while (level.items.len > 0) {
                for (level.items) |subset| {
                    var transitions = try AutoHashMap(u8, AutoHashMap(usize, void)).init(allocator, &.{}, &.{});
                    defer {
                        for (transitions.values()) |*set|
                            set.deinit(allocator);
                        transitions.deinit(allocator);
                    }
                    for (subset) |state| {
                        for (self.states.items[state].out.items) |transition| {
                            if (transition.char) |char| {
                                if (transitions.getPtr(char)) |set| {
                                    try set.put(allocator, transition.to, {});
                                } else {
                                    try transitions.put(allocator, char, try AutoHashMap(usize, void).init(allocator, &.{transition.to}, &.{{}}));
                                }
                            }
                        }
                    }
                    var iter = transitions.iterator();
                    while (iter.next()) |e| {
                        var buf: [256]usize = undefined;
                        var i: usize = 0;
                        for (e.value_ptr.keys()) |state| {
                            buf[i] = state;
                            i += 1;
                        }
                        const reached_states = try self.hull(buf[0..i], subsets.allocator());
                        std.sort.pdq(usize, reached_states, {}, std.sort.asc(usize));
                        const gop = try nfa_to_dfa.getOrPut(allocator, reached_states);
                        if (!gop.found_existing) {
                            var v = self.states.items[reached_states[0]].value;
                            for (reached_states[1..]) |s|
                                v = merge_values(v, self.states.items[s].value);
                            const dfa_state = try dfa.addState(contains(reached_states, sub.end), v);
                            gop.value_ptr.* = dfa_state;
                            try next_level.append(allocator, reached_states);
                        }
                        try dfa.addTransition(nfa_to_dfa.get(subset).?, e.key_ptr.*, gop.value_ptr.*);
                    }
                }
                level.clearRetainingCapacity();
                const old_level = level;
                level = next_level;
                next_level = old_level;
            }
            return dfa;
        }
    };
}
