const std = @import("std");

const NFA = @import("nfa.zig").NFA(void);
const DFA = @import("dfa.zig").DFA(void);

const allocator = std.testing.allocator;

fn initNFA() NFA {
    return NFA.init(allocator, {});
}

fn nothing(_: void, _: void) void {
    return {};
}

const expect = std.testing.expect;

test "optional" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.opt(try nfa.str("test"));
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("test"));
    try expect(dfa.accepts(""));
    try expect(!dfa.accepts("t"));
}

test "repetition" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.rep(try nfa.str("test"));
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("test"));
    try expect(dfa.accepts("testtest"));
    try expect(!dfa.accepts(""));
    try expect(!dfa.accepts("t"));
    try expect(!dfa.accepts("testt"));
}

test "kleene" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.kleene(try nfa.str("test"));
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("testtesttest"));
    try expect(dfa.accepts("test"));
    try expect(dfa.accepts(""));
    try expect(!dfa.accepts("t"));
    try expect(!dfa.accepts("testyay"));
}

test "concat" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.concat(&.{ try nfa.str("hello"), try nfa.str("world") });
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("helloworld"));
    try expect(!dfa.accepts(""));
    try expect(!dfa.accepts("hello"));
    try expect(!dfa.accepts("world"));
}

test "choice" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.choice(&.{ try nfa.str("hi"), try nfa.str("hello") });
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("hi"));
    try expect(dfa.accepts("hello"));
    try expect(!dfa.accepts(""));
    try expect(!dfa.accepts("h"));
    try expect(!dfa.accepts("helloworld"));
}

test "putting it all together" {
    var nfa = initNFA();
    defer nfa.deinit();
    const sub = try nfa.choice(&.{
        try nfa.concat(&.{
            try nfa.rep(try nfa.str("ab")),
            try nfa.str("cd"),
        }),
        try nfa.kleene(try nfa.str("ef")),
    });
    var dfa = try nfa.toDFA(sub, nothing);
    defer dfa.deinit();
    try expect(dfa.accepts("abcd"));
    try expect(dfa.accepts("abababcd"));
    try expect(dfa.accepts("efef"));
    try expect(dfa.accepts(""));
    try expect(!dfa.accepts("abcdef"));
    try expect(!dfa.accepts("g"));
    try expect(!dfa.accepts("a"));
}

const origRegexToDFA = @import("ex.zig").regexToDFA;
fn regexToDFA(regex: []const u8) !DFA {
    return origRegexToDFA(allocator, regex);
}

test "simple regex" {
    var dfa = try regexToDFA("(ab)+cd|(ef)*");
    defer dfa.deinit();
    try expect(dfa.accepts("abcd"));
    try expect(dfa.accepts("abababcd"));
    try expect(dfa.accepts("efef"));
    try expect(dfa.accepts(""));
    try expect(!dfa.accepts("abcdef"));
    try expect(!dfa.accepts("g"));
    try expect(!dfa.accepts("a"));
}

test "charset" {
    var dfa = try regexToDFA("[a%]|b]");
    defer dfa.deinit();
    try expect(dfa.accepts("a"));
    try expect(dfa.accepts("b"));
    try expect(dfa.accepts("]"));
    try expect(dfa.accepts("|"));
    try expect(!dfa.accepts("c"));
}

test "charset range" {
    var dfa = try regexToDFA("[a-c]");
    defer dfa.deinit();
    try expect(dfa.accepts("a"));
    try expect(dfa.accepts("b"));
    try expect(!dfa.accepts("d"));
}

test "charset complement" {
    var dfa = try regexToDFA("[^ab]");
    defer dfa.deinit();
    try expect(!dfa.accepts("a"));
    try expect(!dfa.accepts("b"));
    try expect(dfa.accepts("c"));
}
