const std = @import("std");

const NFA = @import("reg/nfa.zig").NFA(StateInfo);
const DFA = @import("reg/dfa.zig").DFA(StateInfo);
const RegexReader = @import("reg/ex.zig").RegexReader(StateInfo);
const allocator = std.heap.c_allocator;

pub const TokenType = enum(u8) {
    keyword,
    identifier,
    quoted,
    number,
    sym,
    pp,
    comment,
    ignore,
};

pub const Token = struct {
    kind: TokenType,
    len: usize,
};

const StateInfo = struct {
    tt: TokenType,
    acc: bool,
};

fn mergeTokens(a: StateInfo, b: StateInfo) StateInfo {
    if (a.acc != b.acc) {
        return if (a.acc) a else b;
    }
    return .{
        .tt = @enumFromInt(@min(@intFromEnum(a.tt), @intFromEnum(b.tt))),
        .acc = a.acc,
    };
}

pub const RegexTokenizer = struct {
    const Self = @This();
    dfa: DFA,
    pub fn init(comptime rules: []const struct { TokenType, []const u8 }) !Self {
        var nfa = NFA.init(allocator, .{ .tt = .ignore, .acc = false });
        defer nfa.deinit();
        var subs: [rules.len]NFA.Sub = undefined;
        var i: usize = 0;
        while (i < rules.len) : (i += 1) {
            nfa.value = .{ .tt = rules[i][0], .acc = false };
            var rr = RegexReader.init(&nfa, rules[i][1]);
            subs[i] = try rr.read();
            nfa.states.items[subs[i].end].value.acc = true; // HACK
        }
        const sub = try nfa.choice(&subs);
        return .{
            .dfa = try nfa.toDFA(sub, mergeTokens),
        };
    }
    // no need to deinit
    pub fn nextToken(self: Self, buf: []const u8) Token {
        var state: usize = 0;
        var len: usize = 0;
        while (len < buf.len) : (len += 1) {
            const c = buf[len];
            if (self.dfa.getTransition(state, c)) |next_state| {
                state = next_state;
            } else {
                return .{
                    .kind = self.dfa.states.items[state].value.tt,
                    .len = @max(len, 1), // HACK
                };
            }
        }
        return .{
            .kind = self.dfa.states.items[state].value.tt,
            .len = @max(len, 1),
        };
    }
};

// Two interesting optimizations are possible here:
// - We could retokenize only the affected range.
// - We could use a clever data structure (e.g. B-trees augmented with sums of token lengths)
// All in all this would in theory allow updates in O(edit size) time.
// But we don't have working B-trees so we don't and this is O(text size) on every edit. Urgh.
pub const Tokenization = struct {
    const Self = @This();
    tokens: std.ArrayList(Token),
    pub fn init(text: []const u8, tokenizer: RegexTokenizer) !Self {
        var i: usize = 0;
        var tokens = std.ArrayList(Token).init(allocator);
        while (i < text.len) {
            const t = tokenizer.nextToken(text[i..]);
            try tokens.append(t);
            i += t.len;
        }
        return .{ .tokens = tokens };
    }
    pub fn deinit(self: *Self) void {
        self.tokens.deinit();
    }
    pub const Hint = struct {
        ti: usize = 0,
        sum: usize = 0,
    };
    pub fn getPreviousTokenStart(self: Self, i: usize) usize {
        var sum: usize = 0;
        var res: usize = 0;
        var j: usize = 0;
        while (sum < i and j < self.tokens.items.len) : (j += 1) {
            const token = self.tokens.items[j];
            if (token.kind != .ignore)
                res = sum;
            sum += token.len;
        }
        return res;
    }
    pub fn getNextTokenEnd(self: Self, i: usize) usize {
        var sum: usize = 0;
        var j: usize = 0;
        while (sum < i and j < self.tokens.items.len) {
            sum += self.tokens.items[j].len;
            j += 1;
        }
        while (j < self.tokens.items.len) : (j += 1) {
            sum += self.tokens.items[j].len;
            if (self.tokens.items[j].kind != .ignore) break;
        }
        return sum;
    }
    pub fn getToken(self: Self, i: usize, hint: *Hint) Token {
        while (hint.sum <= i and hint.ti < self.tokens.items.len) {
            hint.sum += self.tokens.items[hint.ti].len;
            hint.ti += 1;
        }
        return self.tokens.items[hint.ti - 1];
    }
};
