const std = @import("std");
const fs = std.fs;
const Path = []const u8;

const stdio = @cImport({
    @cInclude("stdio.h");
});

const Pos = @import("pos.zig");

const char = u8;

const allocator = std.heap.c_allocator;
const Allocator = std.mem.Allocator;

const ArrayList = std.ArrayList;

const Tokenization = @import("tok.zig").Tokenization;

pub const FileType = enum {
    plain,
    c,
    fn init(path: Path) FileType {
        return if (std.mem.endsWith(u8, path, ".c")) .c else .plain;
    }
};

fn tokenize(t: FileType, text: []const u8) !Tokenization {
    return Tokenization.init(text, switch (t) {
        .plain => try @import("plain.zig").tokenizer(),
        .c => try @import("c.zig").tokenizer(),
    });
}

pub const TextBuffer = struct {
    const Self = @This();
    chars: ArrayList(char),
    lineOffsets: ArrayList(usize), // where does the respective line end?
    fileType: FileType,
    tokens: Tokenization,
    pub fn empty() !Self {
        var chars = try ArrayList(char).initCapacity(allocator, 1);
        try chars.append('\n');
        return Self{
            .chars = chars,
            .lineOffsets = try calculateLineOffsets(chars.items),
            .tokens = try tokenize(.plain, chars.items),
            .fileType = .plain,
        };
    }
    pub fn open(path: Path) !Self {
        const file = try fs.cwd().openFile(path, .{ .mode = .read_only });
        const reader = file.reader();
        var chars = ArrayList(char).init(allocator);
        {
            defer file.close();
            try reader.readAllArrayList(&chars, 16384);
        }
        const ft = FileType.init(path);
        return .{
            .chars = chars,
            .lineOffsets = try Self.calculateLineOffsets(chars.items),
            .tokens = try tokenize(ft, chars.items),
            .fileType = ft,
        };
    }
    pub fn save(self: *TextBuffer, path: Path) !void {
        const old_ft = self.fileType;
        self.fileType = FileType.init(path);
        if (old_ft != self.fileType) {
            self.tokens.deinit();
            self.tokens = try tokenize(self.fileType, self.chars.items);
        }

        const dirname = fs.path.dirname(path);
        const basename = fs.path.basename(path);
        const tmpname = try std.mem.concat(allocator, u8, &.{ ".~sped-tmp-", basename });
        defer allocator.free(tmpname);
        const tmpath = try fs.path.join(allocator, if (dirname) |name| &.{ name, basename } else &.{basename});
        defer allocator.free(tmpath);
        const file = try fs.cwd().createFile(tmpath, .{});
        {
            defer file.close();
            try file.writeAll(self.chars.items);
        }
        try fs.cwd().rename(tmpath, path);
    }
    pub fn deinit(self: *Self) void {
        self.chars.deinit();
        self.lineOffsets.deinit();
        self.tokens.deinit();
    }
    /// Number of rows, always >= 1.
    pub fn nRows(self: Self) usize {
        return self.lineOffsets.items.len;
    }
    pub fn nCols(self: Self, row: usize) usize {
        return if (row == 0)
            self.lineOffsets.items[0]
        else
            self.lineOffsets.items[row] - self.lineOffsets.items[row - 1] - 1;
    }
    // TODO LFs (<=)
    pub fn inBounds(self: Self, pos: Pos) bool {
        return pos.row < self.nRows() and pos.col < self.nCols(pos.row);
    }
    fn rowStart(self: Self, row: usize) usize {
        return if (row == 0) 0 else self.lineOffsets.items[row - 1] + 1;
    }
    pub fn getIdx(self: Self, pos: Pos) usize {
        return self.rowStart(pos.row) + pos.col;
    }
    fn loLt(_: void, lhs: usize, rhs: usize) bool {
        return lhs < rhs;
    }
    /// O(log n)
    pub fn getPos(self: Self, idx: usize) Pos {
        const row = std.sort.lowerBound(usize, idx, self.lineOffsets.items, void{}, loLt);
        return .{ .row = row, .col = idx - self.rowStart(row) };
    }
    pub fn get(self: Self, pos: Pos) ?char {
        return if (self.inBounds(pos)) self.chars.items[self.getIdx(pos)] else null;
    }
    pub fn getRange(self: Self, from: usize, to: usize) []const u8 {
        return self.chars.items[from..to];
    }
    fn calculateLineOffsets(chars: []const u8) !ArrayList(usize) {
        var lineOffsets = ArrayList(usize).init(allocator);
        var i: usize = 0;
        while (std.mem.indexOfScalar(char, chars[i..], '\n')) |offset| {
            try lineOffsets.append(i + offset);
            i += offset + 1;
        }
        if (chars.len == 0 or i < chars.len) // no trailing newline
            try lineOffsets.append(chars.len); // note: no newline!
        return lineOffsets;
    }
    pub fn replaceRange(self: *Self, from: usize, len: usize, replacement: []const u8) !void {
        try self.chars.replaceRange(from, len, replacement);
        self.lineOffsets.deinit();
        self.lineOffsets = try calculateLineOffsets(self.chars.items);
        self.tokens.deinit();
        self.tokens = try tokenize(self.fileType, self.chars.items);
    }
};
